package cz.jzitnik.gameoflife.game;

/**
 * The GameOfLife class represents the logic for Conway's Game of Life.
 */
public class GameOfLife {
    private boolean[][] board;
    private final int rows;
    private final int cols;
    private static final int[] DIRECTIONS = {-1, 0, 1};

    /**
     * Constructs a new GameOfLife instance with the given number of rows and columns.
     *
     * @param rows Number of rows in the game board.
     * @param cols Number of columns in the game board.
     */
    public GameOfLife(int rows, int cols) {
        this.rows = rows;
        this.cols = cols;
        this.board = new boolean[rows][cols];
    }

    /**
     * Advances the game by one generation according to Conway's Game of Life rules.
     */
    public void nextGeneration() {
        boolean[][] newBoard = new boolean[rows][cols];

        for (int i = 0; i < rows; i++) {
            for (int j = 0; j < cols; j++) {
                int liveNeighbors = countLiveNeighbors(i, j);
                // A live cell survives if it has 2 or 3 live neighbors, otherwise it dies.
                // A dead cell becomes alive if it has exactly 3 live neighbors.
                newBoard[i][j] = (board[i][j] && (liveNeighbors == 2 || liveNeighbors == 3)) || (!board[i][j] && liveNeighbors == 3);
            }
        }
        board = newBoard;
    }

    /**
     * Counts the number of live neighbors for a given cell.
     *
     * @param row The row index of the cell.
     * @param col The column index of the cell.
     * @return The number of live neighbors.
     */
    private int countLiveNeighbors(int row, int col) {
        int count = 0;
        for (int dr : DIRECTIONS) {
            for (int dc : DIRECTIONS) {
                if (dr == 0 && dc == 0) continue;
                int newRow = row + dr, newCol = col + dc;
                if (isValid(newRow, newCol) && board[newRow][newCol]) count++;
            }
        }
        return count;
    }

    /**
     * Checks if the given row and column indices are within the valid range of the board.
     *
     * @param row The row index.
     * @param col The column index.
     * @return True if the indices are valid, false otherwise.
     */
    private boolean isValid(int row, int col) {
        return row >= 0 && row < rows && col >= 0 && col < cols;
    }

    /**
     * Toggles the state of a specific cell between alive and dead.
     *
     * @param row The row index of the cell.
     * @param col The column index of the cell.
     */
    public void toggleCell(int row, int col) {
        if (isValid(row, col)) {
            board[row][col] = !board[row][col];
        }
    }

    /**
     * Checks if the entire board is empty (i.e., all cells are dead).
     *
     * @return True if no cells are alive, false otherwise.
     */
    public boolean isBoardEmpty() {
        for (boolean[] row : board) {
            for (boolean cell : row) {
                if (cell) return false;
            }
        }
        return true;
    }

    /**
     * Retrieves the state of a specific cell.
     *
     * @param row The row index of the cell.
     * @param col The column index of the cell.
     * @return True if the cell is alive, false if it is dead or out of bounds.
     */
    public boolean getCell(int row, int col) {
        return isValid(row, col) && board[row][col];
    }
}
